// Author: Theodore Norvell
// Original version: June 2004

#ifndef FUNCTION_LL_H
#define FUNCTION_LL_H

#include <cassert>
#include <new>

using namespace std ;

// Function
///////////////

// A class for representing finite parial functions.
// Restriction: The domain type must support a == and a != operator
// Abstract data:  F a partial function represented as a set of pairs.

template <class DomainT, class RangeT>
class Function {

  public:
  // Exceptions

	class SpaceException { } ;

  // Creators

	Function(void) ;
	// (void Constructor) Create an empty function.
	// Post: F' = {}
	// May throw SpaceException

	Function(const Function &src) ;
	// (Copy constructor).
	// Post: F'=src.F
	// May throw SpaceException.

  // Destroyers

	~Function(void) ;
	// Destroy this function.

  // Accessors

	bool inDomain(const DomainT &d) const ;
	// Report whether d is in the preimage of this function
	// Post: Result = ( d in preimage(F) ) and F' = F

	RangeT at(const DomainT &d) const ;
	// Apply this function to d.
	// Pre:  d in preimage(F)
	// Post: Result = F(d) and F' = F

    RangeT &atRef(const DomainT &d) ;
    // Apply this function to d be return a reference that may
    // be modified.
    // Pre: d in preimage(F)
    // Post: Result = a reference to F(d)
    // Note: The returned reference becomes invalid as soon as any
    // mutator is called for this object.

	bool empty(void) const ;
	// Report whether this function is empty.
	// Post: Result = ( F={} ) and F' = F

	void getAny(DomainT &d, RangeT &r) const ;
	// Return any pair associated by this function.
	// Pre: F != {}
	// Post: (d',r') in F and F' = F

    bool operator==( const Function &other ) const ;

    bool operator!=( const Function &other ) const ;

  // Mutators:
	const Function& operator=(const Function &src) ;
	// (Assignment operator)
	// Post: F'=src.F and Result= a reference to src.
	// May throw SpaceException.

	void add(const DomainT &d, const RangeT &r ) ;
	// Add d to the preimage of this function so that d maps to r.
	// Pre: not d in preimage(F)
	// Post: F' = F U (d, r)
	// May throw SpaceException.

	void remove( const DomainT &d ) ;
	// Removes d from the preimage of this function
	// Pre: d in preimage(F)
	// Post: F' = {(a,b) in F | a != d }

  private:
    // Representation: The set of pairs is represented by an unordered,
    // null-terminated, singly-linked list of Nodes.
    struct Node {
    	DomainT domainElem ;
    	RangeT  rangeElem ;
    	Node *next ;
    } ;
    
    Node * head ;
    
    Node const* findNode( DomainT const & domainElem ) const ;
    Node ** findLink( DomainT const & domainElem ) ;
    int length( ) const ;  
} ;

///////////////////////////////////////
//
//  IMPLEMENTATION
//
///////////////////////////////////////

/***************************
 * Public member functions
 ***************************/
 
 
// Default Constructor. Create an empty function.
// Post: F' = {}
// May throw SpaceException
template<class DomainT, class RangeT>
Function<DomainT, RangeT>::Function(void)
{
	head = 0 ;
}

// Copy constructor
// Post: F'=src.F
// May throw SpaceException.
template<class DomainT, class RangeT>
Function<DomainT, RangeT>::	Function(const Function &src) {
	head = 0 ;
	Node *p = src.head ;
	while( p != 0 ) {
	add( p->domainElem, p->rangeElem ) ;
		p = p->next ; }
}

// Destructor
template<class DomainT, class RangeT>
Function<DomainT, RangeT>::	~Function(void) {
	while( head != 0 ) {
		remove( head->domainElem ) ; }
}

// inDomain  Report whether d is in the preimage of this function
// NOTE!  I really named this one badly!!!!
// Post: Result = ( d in preimage(F) ) and F' = F
template<class DomainT, class RangeT>
bool Function<DomainT, RangeT>:: inDomain(const DomainT &d) const {
	return findNode( d ) != 0 ;
}

// at  Apply this function to d.
// Pre:  d in preimage(F)
// Post: Result = F(d) and F' = F
template<class DomainT, class RangeT>
RangeT Function<DomainT, RangeT>:: at(const DomainT &d) const {
	Node const * p = findNode( d ) ;
	assert( p != 0 ) ;
	return p->rangeElem ;
}


// atRef Apply this function to d be return a reference that may
// be modified.
// Pre: d in preimage(F)
// Post: Result = a reference to F(d)
// Note: The returned reference becomes invalid as soon as any
// mutator is called for this object.

template<class DomainT, class RangeT>
RangeT& Function<DomainT, RangeT>:: atRef(const DomainT &d) {
	Node **linkP = findLink( d ) ;
	assert( *linkP != 0 ) ;
	return (*linkP)->rangeElem ;
}

// empty  Report whether this function is empty.
// Post: Result = ( F={} ) and F' = F
template<class DomainT, class RangeT>
bool Function<DomainT, RangeT>:: empty(void) const {
	return head == 0 ;
}

// Return any pair associated by this function.
// Pre: F != {}
// Post: (d',r') in F and F' = F
template<class DomainT, class RangeT>
void Function<DomainT, RangeT>:: getAny(DomainT &d, RangeT &r) const {
	assert( head != 0 ) ;
	d = head->domainElem ;
	r = head->rangeElem ;
}

// ==
template<class DomainT, class RangeT>
bool Function<DomainT, RangeT>:: operator==( const Function &other ) const {
	if( length() != other.length() ) return false ;
	Node * p = head ;
	while( p != 0 ) {
		if( ! other.inDomain( p->domainElem ) ) return false ;
		if( p->rangeElem != other.at( p->domainElem ) ) return false ;
		p = p->next ; }
	return true ;
}

// !=
template<class DomainT, class RangeT>
bool Function<DomainT, RangeT>:: operator!=( const Function &other ) const {
	return !( *this == other ) ;
}

// (Assignment operator)
// Post: F'=src.F and Result= a reference to src.
// May throw SpaceException.
template<class DomainT, class RangeT>
const Function<DomainT, RangeT>& Function<DomainT, RangeT>:: operator=(const Function &src) {
	if( this != & src ) {
		while( head != 0 ) { remove( head->domainElem ) ; }
	    Node *p = src.head ;
	    while( p != 0 ) {
		    add( p->domainElem, p->rangeElem ) ;
			p = p->next ; } }
	return src ;
}

// add(d, r) Add d to the preimage of this function so that d maps to r.
// Pre: not d in preimage(F)
// Post: F' = F U (d, r)
// May throw SpaceException.
template<class DomainT, class RangeT>
void Function<DomainT, RangeT>:: add(const DomainT &d, const RangeT &r ) {
	assert( ! inDomain( d ) ) ;
	Node *temp = new(nothrow) Node ;
	if( temp == 0 ) throw SpaceException() ;
	temp->domainElem = d ;
	temp->rangeElem = r ;
	temp->next = head ;
	head = temp ;

}

// remove( d ) Removes d from the preimage of this function
// Pre: d in preimage(F)
// Post: F' = {(a,b) in F | a != d }
template<class DomainT, class RangeT>
void Function<DomainT, RangeT>:: remove( const DomainT &d ) {
	Node **linkP = findLink( d ) ;
	assert (*linkP != 0 ) ;
	Node *temp = *linkP ;
	*linkP = (*linkP)->next ;
	delete temp ;
}

/***************************
 * Public member functions
 ***************************/


// findNode(domianElem)  -- return a pointer to the node with domainElem
// Post: *  result==0, if no node has domainElem, and
//       *  result points to a node having domainElem, otherwise
         
template<class DomainT, class RangeT>
Function<DomainT, RangeT>::Node const*
Function<DomainT, RangeT>:: findNode( DomainT const & domainElem ) const {
	Node *p = head ;
	while( p != 0 && p->domainElem != domainElem ) {
		p = p->next ; }
	return p ;
}

// findLink( domainElem ) Return a pointer to the link pointing to the
//                        node with domainElem
//
// Post: *  result points to some link in the list
//       *  *result==0, if no node has domainElem, and
//       *  *result points to a node having domainElem, otherwise

template<class DomainT, class RangeT>
Function<DomainT, RangeT>::Node **
Function<DomainT, RangeT>:: findLink( DomainT const & domainElem ) {
	Node **linkP = & head ;
	while( *linkP != 0 && (*linkP)->domainElem != domainElem ) {
		linkP = & (*linkP)->next ; }
	return linkP ;
}


// length -- returns the length of the list

template<class DomainT, class RangeT>
int Function<DomainT, RangeT>:: length() const {
	int len = 0 ;
	Node *p = head ;
	while( p != 0  ) {
		++len, p = p->next ;
	}
	return len ;
}

#endif